/*

  CompineRotors (c)2015,2016 Hardy Lau, DL1GLH

  Das Programm stellt einen virtuellen ARS-USB mit Azimut und Elevation
  bereit und nutzt dabei einen ARS-USB fuer Azimut und einen M2 RC2800 fuer
  die Elevation.

  Das Programm oeffnet zwei serielle Kanaele fuer ARS-USB und M2 RC2800
  und stellt ein Pseudo-TTY fuer ser2net zur Verfuegung.

  ARS-USB, M2 RC2800, BananaPi, Seriell-USB-Wandler
  ARSVCOM 6.3.4 mit Update im Sekundenrhytmus
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#define _XOPEN_SOURCE 
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/termios.h>
#include <math.h>
 
int isdaemon = 0;  //Laeuft als daemon. Keine Ausgaben

int pid;           //PID des Prozesses
int pidfile;       //PID-File fuer systemd

int pseudo;        //Pseudo-TTY mit emuliertem ARS-USB AzEl
char pseudolink[] = "/tmp/emul-ars";
char pseudoname[100];
int ars, m2;       //Handle der Rotorsteuerungen

//Aufraeumen bei Programmende oder Fehler
//
void signal_catcher( int signal)
{
 close( pseudo);
 unlink( pseudolink);
 close( ars);
 close( m2);
 unlink( "/var/run/cr.pid");
 exit( 0);
 sleep(1);
}

//Eine Zeile mit Timeout lesen
//
int lesezeile( int fd, char* buffer, int sizeofbuffer)
{
 int i = 0;
 char ss[1];
 int gelesen;
 fd_set rfds;
 struct timeval tv;
 int retval;

 while( i < sizeofbuffer) {
   FD_ZERO(&rfds);
   FD_SET(fd, &rfds);
   tv.tv_sec = 0;
   tv.tv_usec = 100000L;
   retval = select(fd+1, &rfds, NULL, NULL, &tv);

   if(retval) {
     gelesen = read( fd, ss, 1);
     if( gelesen <= 0) return( -1); 
     buffer[i++] = ss[0];
     buffer[i] = '\0';
     if( ss[0] == '\n') return( i);
   } else { //Timeout!
      return( i);
   }

 }
}

//Hauptprogramm
//
main( int argc, char* argv[])
{
 int res;
 char zeile[100]; 
 char ausgabe[100];
 int arsleseversuche = 0;

 int leerlaufzaehler = 0;
 int bereitsverbunden = 0;

 //M2 RC2800 nimmt waeherend Bewegung keine neuen Kommandos an
 //Kommnado merken und spaeter bei C2-Kommando wiederholen
 char letzte_m2_zeile[100];
 int n_kommando_wartet = 0;
 int elevation_m2_wartet = 0;

 int anz;
 struct termios term;

 int azimut = 0, elevation = 0;
 float elevation_float = 0.0;

 strcpy( letzte_m2_zeile, "");

 //fork
 //
 if( argc == 1) { //Kein Argument also deamon
   pid = fork();
   if( pid) { //Parent
     //run-pid-file anlegen
     unlink( "/var/run/cr.pid");
     pidfile = open( "/var/run/cr.pid", O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
     sprintf( zeile, "%d", pid);
     write( pidfile, zeile, strlen( zeile));
     close( pidfile);
     exit( 0);
   } else {
   //Child
     close( 0);
     close( 1);
     close( 2);
     chdir( "/");
     isdaemon = 1;
   }
 }

 //Signal-Handler fuer Programmende anegen
 //
 signal( SIGHUP, SIG_IGN); 
 signal( SIGTERM, signal_catcher);
 
 //Pseuto-TTY generieren und Link erzeugen
 //
 pseudo = open( "/dev/ptmx", O_RDWR);
 if( pseudo <= 0) if( !isdaemon) printf( "Error open /dev/ptmx\n");
 res = grantpt( pseudo);
 if( res) if( !isdaemon) printf( "Error grantpt\n");
 res = unlockpt( pseudo);
 if( res) if( !isdaemon) printf( "Error unlockpt\n");
 strcpy( pseudoname,  (char*)ptsname(pseudo));
 if( !isdaemon) printf( "%s\n", pseudoname);
 unlink( pseudolink);
 symlink( pseudoname, pseudolink);

 //TTY fuer ARS-USB oeffnen
 //
 ars = open( "/dev/ttyACM0", O_RDWR);
 if( ars <= 0) ars = open( "/dev/ttyACM1", O_RDWR);
 if( ars <= 0) signal_catcher( 0);
 term.c_iflag = IGNBRK | INPCK;
 term.c_oflag = 0;
 term.c_cflag = CS8 | CREAD;
 term.c_lflag = 0;
 cfsetospeed( &term, B38400);
 cfsetispeed( &term, B38400);
 if( tcsetattr( ars, TCSANOW, &term))
   signal_catcher( 0);

 //TTY fuer M2 RC2800 oeffnen
 //
 m2 = open( "/dev/ttyUSB0", O_RDWR);
 if( m2 <= 0) m2 = open( "/dev/ttyUSB1", O_RDWR); 
 if( m2 <= 0) signal_catcher( 0);
 term.c_iflag = IGNBRK | INPCK;
 term.c_oflag = 0;
 term.c_cflag = CS8 | CREAD;
 term.c_lflag = 0;
 cfsetospeed( &term, B9600);
 cfsetispeed( &term, B9600);
 if( tcsetattr( m2, TCSANOW, &term))
   signal_catcher( 0);

 //Kommandointerpreter (Schleife)
 //

 while(1) {
   //Naechstes Kommendo lesen
   //
   anz = lesezeile( pseudo, zeile, sizeof(zeile));
   if( anz < 0) signal_catcher( 0);
   zeile[anz] = '\0';

   //Kein neues Kommando, bei Timeout (Verbindungsverlusst) beenden
   //
   if( anz == 0) {
     if( bereitsverbunden && leerlaufzaehler++ > 50) signal_catcher( 0);
     continue;
   }
   leerlaufzaehler = 0;

   //Neues Kommando bearbeiten
   //
   if( !isdaemon) printf( "KOMMANDO: %s \n", zeile);

   //"O"
   if( !strncmp( "O", zeile, 1)) {
     write( pseudo, "RCI-USB by EA4TX\r\nVer2.1k Firm:111\r\n", sizeof("RCI-USB by EA4TX\r\nVer2.1k Firm:111\r\n"));
     if( write( m2, "E\r\n", strlen( "E\r\n")) <= 0) signal_catcher( 0);
     bereitsverbunden = 1;
     continue;
   }

   //"C2"
   if( !strncmp( "C2", zeile, 2)) {

     //ARS-USB nach Azimut fragen
     //
     arsleseversuche = 4;
     do {
       anz = lesezeile( ars, zeile, sizeof(zeile));
       if( anz < 0) signal_catcher( 0);
       if( !isdaemon) printf( "ARS-FRAGEN...");
       anz = write( ars, "C2\n", strlen("C2\n"));
       if( anz <= 0) signal_catcher( 0);
       if( !isdaemon) printf( "LESEN...");
       anz = lesezeile( ars, zeile, sizeof(zeile));
       if( anz < 0) signal_catcher( 0);
     } while( anz < 6 && arsleseversuche--);
     if( !arsleseversuche) {
       if( !isdaemon) printf( "ARS TIMEOUT!!!\n");
       signal_catcher( 0);
     }
     if( !isdaemon) printf( "ARS-ANTWORT: %s\n", zeile);
     azimut = 0;
     if( strchr( zeile, '+')) {
       sscanf( strchr( zeile, '+'), "+%04d", &azimut);
       if( !isdaemon) printf( "ARS-ANTWORT als Zahl: %d\n", azimut);
     }

     //M2 RC2800 nach Elevation fragen
     //
     while( lesezeile( m2, zeile, sizeof(zeile)) > 0);
     anz = write( m2, "E\r", strlen( "E\r"));
     if( anz <= 0) signal_catcher( 0);
     anz = lesezeile( m2, zeile, sizeof(zeile));
     if( anz <= 0) signal_catcher( 0);
     zeile[ anz] = '\0';
     if( anz < 5) anz = lesezeile( m2, zeile, sizeof(zeile));
     strcpy( letzte_m2_zeile, zeile);
     if( !isdaemon) printf( "M2-ANTWORT: %s\n", zeile);

     sscanf( zeile, "E=%d.", &elevation);
     elevation = (int)roundf( elevation);
     if( !isdaemon) printf( "M2-ANTWORT als Zahl: %d\n", elevation);

     //M2 eventuell wartendes Kommando senden
     //
     if( n_kommando_wartet && strchr( letzte_m2_zeile, 'S')) {
       if( n_kommando_wartet) {
         if( !isdaemon) printf( "KOMMANDO an M2: %s\n", ausgabe);
         sprintf( ausgabe, "E%d\r", elevation_m2_wartet);
         write( m2, ausgabe, strlen( ausgabe));
       }
       n_kommando_wartet--;
     }

     //Antworten
     //
     sprintf( ausgabe, "+%04d+%04d\r\n", azimut, elevation);
     anz = write( pseudo, ausgabe, strlen(ausgabe));
     if( anz <= 0) signal_catcher( 0);
     continue;
   }

   //"M"
   if( zeile[0] == 'M') {
     if( sscanf( zeile, "M%03d", &azimut) == 1) {
       sprintf( ausgabe, "M%d\r", azimut);
       write( ars, ausgabe, strlen(ausgabe));
     }
   }

   //"N"
   if( zeile[0] == 'N') {
     if( sscanf( zeile, "N%03d", &elevation) == 1) {
       sprintf( ausgabe, "E%d\r", elevation);
       //Testen ob Rotor steht. Nimmt dann keine Kommandos
       elevation_m2_wartet = elevation;
       if( strchr( letzte_m2_zeile, 'M')) {
         //Stopp senden
         write( m2, "S\r\n", strlen( "S\r\n"));
         n_kommando_wartet = 9;
       } else {
         if( !isdaemon) printf( "KOMMANDO an M2: %s\n", ausgabe);
         write( m2, ausgabe, strlen( ausgabe));
       }
     }
     continue;
   }
 
   //"W"
   if( zeile[0] == 'W') {
     if( sscanf( zeile, "W%03d %03d", &azimut, &elevation) == 2) {
       sprintf( zeile, "M%03d\n", azimut);
       write( ars, zeile, strlen(zeile));
       sprintf( zeile, "E%d.0\r\n", elevation);
       //Testen ob Rotor steht. Nimmt dann keine Kommandos
       elevation_m2_wartet = elevation;
       if( strchr( letzte_m2_zeile, 'M')) {
         //Stopp senden
         write( m2, "S\r\n", strlen( "S\r\n"));
         n_kommando_wartet = 9;
       } else {
         if( !isdaemon) printf( "KOMMANDO an M2: %s\n", ausgabe);
         write( m2, zeile, strlen( zeile));
       }
     }
     continue;
   } 

   //"A"
   if( zeile[0] == 'A') {
     write( ars, "A\n", strlen( "A\n"));
     continue;
   }

   //"E"
   if( zeile[0] == 'E') {
     write( m2, "S\r\n", strlen( "S\r\n"));
     continue;
   }
  
   //"S"
   if( zeile[0] == 'S') {
     write( ars, "A\n", strlen( "A\n"));
     write( m2, "S\r\n", strlen( "S\r\n"));
     continue;
   }

 }
}
